"""
Extra.

pymdown.extra
A wrapper that emulate PHP Markdown Extra.
Re-packages Python Markdowns 'extra' extensions,
but substitutes a few extensions with PyMdown extensions:

- fenced_code --> superfences
- smartstrong --> betterem

MIT license.

Copyright (c) 2015 - 2017 Isaac Muse <isaacmuse@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from markdown import Extension

extra_extensions = [
    'pymdownx.betterem',
    'pymdownx.superfences',
    'markdown.extensions.footnotes',
    'markdown.extensions.attr_list',
    'markdown.extensions.def_list',
    'markdown.extensions.tables',
    'markdown.extensions.abbr',
    'markdown.extensions.md_in_html'
]

extra_extension_configs = {}


class ExtraExtension(Extension):
    """Add various extensions to Markdown class."""

    def __init__(self, *args, **kwargs):
        """Initialize."""

        self.config = kwargs.pop('configs', {})
        self.config.update(extra_extension_configs)
        self.config.update(kwargs)

    def extendMarkdown(self, md):
        """Register extension instances."""

        md.registerExtensions(extra_extensions, self.config)


def makeExtension(*args, **kwargs):
    """Return extension."""

    return ExtraExtension(*args, **kwargs)



# -*- coding: utf-8 -*-
"""English US keymap."""

keymap = {
    # Digits
    "0": "0",
    "1": "1",
    "2": "2",
    "3": "3",
    "4": "4",
    "5": "5",
    "6": "6",
    "7": "7",
    "8": "8",
    "9": "9",

    # Letters
    "a": "A",
    "b": "B",
    "c": "C",
    "d": "D",
    "e": "E",
    "f": "F",
    "g": "G",
    "h": "H",
    "i": "I",
    "j": "J",
    "k": "K",
    "l": "L",
    "m": "M",
    "n": "N",
    "o": "O",
    "p": "P",
    "q": "Q",
    "r": "R",
    "s": "S",
    "t": "T",
    "u": "U",
    "v": "V",
    "w": "W",
    "x": "X",
    "y": "Y",
    "z": "Z",

    # Space
    "space": "Space",

    # Punctuation
    "backslash": "\\",
    "bar": "|",
    "brace-left": "{",
    "brace-right": "}",
    "bracket-left": "[",
    "bracket-right": "]",
    "colon": ":",
    "comma": ",",
    "double-quote": "\"",
    "equal": "=",
    "exclam": "!",
    "grave": "`",
    "greater": ">",
    "less": "<",
    "minus": "-",
    "period": ".",
    "plus": "+",
    "question": "?",
    "semicolon": ";",
    "single-quote": "'",
    "slash": "/",
    "tilde": "~",
    "underscore": "_",

    # Navigation keys
    "arrow-up": "Up",
    "arrow-down": "Down",
    "arrow-left": "Left",
    "arrow-right": "Right",
    "page-up": "Page Up",
    "page-down": "Page Down",
    "home": "Home",
    "end": "End",


    # Edit keys
    "backspace": "Backspace",
    "delete": "Del",
    "insert": "Ins",
    "tab": "Tab",

    # Action keys
    "break": "Break",
    "caps-lock": "Caps Lock",
    "clear": "Clear",
    "eject": "Eject",
    "enter": "Enter",
    "escape": "Esc",
    "help": "Help",
    "print-screen": "Print Screen",
    "scroll-lock": "Scroll Lock",

    # Numeric keypad
    "num0": "Num 0",
    "num1": "Num 1",
    "num2": "Num 2",
    "num3": "Num 3",
    "num4": "Num 4",
    "num5": "Num 5",
    "num6": "Num 6",
    "num7": "Num 7",
    "num8": "Num 8",
    "num9": "Num 9",
    "num-asterisk": "Num *",
    "num-clear": "Num Clear",
    "num-delete": "Num Del",
    "num-equal": "Num =",
    "num-lock": "Num Lock",
    "num-minus": "Num -",
    "num-plus": "Num +",
    "num-separator": "Num .",
    "num-slash": "Num /",
    "num-enter": "Num Enter",

    # Modifier keys
    "alt": "Alt",
    "alt-graph": "AltGr",
    "command": "Cmd",
    "control": "Ctrl",
    "function": "Fn",
    "left-alt": "Left Alt",
    "left-command": "Left Command",
    "left-control": "Left Ctrl",
    "left-meta": "Left Meta",
    "left-option": "Left Option",
    "left-shift": "Left Shift",
    "left-super": "Left Super",
    "left-windows": "Left Win",
    "meta": "Meta",
    "option": "Option",
    "right-alt": "Right Alt",
    "right-command": "Right Command",
    "right-control": "Right Ctrl",
    "right-meta": "Right Meta",
    "right-option": "Right Option",
    "right-shift": "Right Shift",
    "right-super": "Right Super",
    "right-windows": "Right Win",
    "shift": "Shift",
    "super": "Super",
    "windows": "Win",

    # Function keys
    "f1": "F1",
    "f2": "F2",
    "f3": "F3",
    "f4": "F4",
    "f5": "F5",
    "f6": "F6",
    "f7": "F7",
    "f8": "F8",
    "f9": "F9",
    "f10": "F10",
    "f11": "F11",
    "f12": "F12",
    "f13": "F13",
    "f14": "F14",
    "f15": "F15",
    "f16": "F16",
    "f17": "F17",
    "f18": "F18",
    "f19": "F19",
    "f20": "F20",
    "f21": "F21",
    "f22": "F22",
    "f23": "F23",
    "f24": "F24",

    # Extra keys
    "backtab": "Back Tab",
    "browser-back": "Browser Back",
    "browser-favorites": "Browser Favorites",
    "browser-forward": "Browser Forward",
    "browser-home": "Browser Home",
    "browser-refresh": "Browser Refresh",
    "browser-search": "Browser Search",
    "browser-stop": "Browser Stop",
    "context-menu": "Menu",
    "copy": "Copy",
    "mail": "Mail",
    "media": "Media",
    "media-next-track": "Next Track",
    "media-pause": "Pause",
    "media-play": "Play",
    "media-play-pause": "Play/Pause",
    "media-prev-track": "Previous Track",
    "media-stop": "Stop",
    "print": "Print",
    "reset": "Reset",
    "select": "Select",
    "sleep": "Sleep",
    "volume-down": "Volume Down",
    "volume-mute": "Mute",
    "volume-up": "Volume Up",
    "zoom": "Zoom",
    "power": "Power",
    "fingerprint": "Fingerprint",

    # Mouse
    "left-button": "Left Button",
    "middle-button": "Middle Button",
    "right-button": "Right Button",
    "x-button1": "X Button 1",
    "x-button2": "X Button 2"
}

aliases = {
    "add": "num-plus",
    "altgr": "alt-graph",
    "apps": "context-menu",
    "back": "backspace",
    "bksp": "backspace",
    "bktab": "backtab",
    "cancel": "break",
    "capital": "caps-lock",
    "close-brace": "brace-right",
    "close-bracket": "bracket-right",
    "clr": "clear",
    "cmd": "command",
    "cplk": "caps-lock",
    "ctrl": "control",
    "dblquote": "double-quote",
    "decimal": "num-separator",
    "del": "delete",
    "divide": "num-slash",
    "down": "arrow-down",
    "esc": "escape",
    "return": "enter",
    "exclamation": "exclam",
    "favorites": "browser-favorites",
    "fn": "function",
    "forward": "browser-forward",
    "grave-accent": "grave",
    "greater-than": "greater",
    "gt": "greater",
    "hyphen": "minus",
    "ins": "insert",
    "lalt": "left-alt",
    "launch-mail": "mail",
    "launch-media": "media",
    "lbutton": "left-button",
    "lcmd": "left-command",
    "lcommand": "left-command",
    "lcontrol": "left-control",
    "lctrl": "left-control",
    "left": "arrow-left",
    "left-cmd": "left-command",
    "left-ctrl": "left-control",
    "lopt": "left-option",
    "loption": "left-option",
    "left-opt": "left-option",
    "left-win": "left-windows",
    "less-than": "less",
    "lmeta": "left-meta",
    "lshift": "left-shift",
    "lsuper": "left-super",
    "lt": "less",
    "lwin": "left-windows",
    "lwindows": "left-windows",
    "mbutton": "middle-button",
    "menu": "context-menu",
    "multiply": "num-asterisk",
    "mute": "volume-mute",
    "next": "page-down",
    "next-track": "media-next-track",
    "num-del": "num-delete",
    "numlk": "num-lock",
    "open-brace": "brace-left",
    "open-bracket": "bracket-left",
    "opt": "option",
    "page-dn": "page-down",
    "page-up": "page-up",
    "pause": "media-pause",
    "pg-dn": "page-down",
    "pg-up": "page-up",
    "pipe": "bar",
    "play": "media-play",
    "play-pause": "media-play-pause",
    "prev-track": "media-prev-track",
    "prior": "page-up",
    "prtsc": "print-screen",
    "question-mark": "question",
    "ralt": "right-alt",
    "rbutton": "right-button",
    "rcontrol": "right-control",
    "rcmd": "right-command",
    "rcommand": "right-command",
    "rctrl": "right-control",
    "refresh": "browser-refresh",
    "right": "arrow-right",
    "right-cmd": "right-command",
    "right-ctrl": "right-control",
    "right-meta": "right-meta",
    "right-opt": "right-option",
    "right-win": "right-windows",
    "rmeta": "right-meta",
    "ropt": "right-option",
    "roption": "right-option",
    "rshift": "right-shift",
    "rsuper": "right-super",
    "rwin": "right-windows",
    "rwindows": "right-windows",
    "scroll": "scroll-lock",
    "search": "browser-search",
    "separator": "num-separator",
    "spc": "space",
    "stop": "media-stop",
    "subtract": "num-minus",
    "tabulator": "tab",
    "up": "arrow-up",
    "vol-down": "volume-down",
    "vol-mute": "volume-mute",
    "vol-up": "volume-up",
    "win": "windows",
    "xbutton1": "x-button1",
    "xbutton2": "x-button2"
}



"""
Smart Symbols.

pymdownx.smartsymbols
Really simple plugin to add support for:
  copyright, trademark, and registered symbols
  plus/minus, not equal, arrows via:

    copyright  = `(c)`
    trademark  = `(tm)`
    registered = `(r)`
    plus/minus = `+/-`
    care/of    = `c/o`
    fractions  = `1/2` etc.
        (only certain available unicode fractions)
    arrows:
        left   = `<--`
        right  = `-->`
        both   = `<-->`
    not equal  = `=/=`
       (maybe this could be =/= in the future as this might be more
        intuitive to non-programmers)

MIT license.

Copyright (c) 2014 - 2017 Isaac Muse <isaacmuse@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from markdown import Extension
from markdown import treeprocessors
from markdown.util import Registry
from markdown.inlinepatterns import HtmlInlineProcessor

RE_TRADE = ("smart-trademark", r'\(tm\)', r'&trade;')
RE_COPY = ("smart-copyright", r'\(c\)', r'&copy;')
RE_REG = ("smart-registered", r'\(r\)', r'&reg;')
RE_PLUSMINUS = ("smart-plus-minus", r'\+/-', r'&plusmn;')
RE_NOT_EQUAL = ("smart-not-equal", r'=/=', r'&ne;')
RE_CARE_OF = ("smart-care-of", r'\bc/o\b', r'&#8453;')
RE_ORDINAL_NUMBERS = (
    "smart-ordinal-numbers",
    r'''(?x)
    \b
    (?P<leading>(?:[1-9][0-9]*)?)
    (?P<tail>(?<=1)(?:1|2|3)th|1st|2nd|3rd|[04-9]th)
    \b
    ''',
    lambda m: '%s%s<sup>%s</sup>' % (
        m.group('leading') if m.group('leading') else '',
        m.group('tail')[:-2], m.group('tail')[1:]
    )
)
RE_ARROWS = (
    "smart-arrows",
    r'(?P<arrows>\<-{2}\>|(?<!-)-{2}\>|\<-{2}(?!-))',
    lambda m: ARR[m.group('arrows')]
)
RE_FRACTIONS = (
    "smart-fractions",
    r'(?<!\d)(?P<fractions>1/4|1/2|3/4|1/3|2/3|1/5|2/5|3/5|4/5|1/6|5/6|1/8|3/8|5/8|7/8)(?!\d)',
    lambda m: FRAC[m.group('fractions')]
)

REPL = {
    'trademark': RE_TRADE,
    'copyright': RE_COPY,
    'registered': RE_REG,
    'plusminus': RE_PLUSMINUS,
    'arrows': RE_ARROWS,
    'notequal': RE_NOT_EQUAL,
    'fractions': RE_FRACTIONS,
    'ordinal_numbers': RE_ORDINAL_NUMBERS,
    'care_of': RE_CARE_OF
}

FRAC = {
    "1/4": "&frac14;",
    "1/2": "&frac12;",
    "3/4": "&frac34;",
    "1/3": "&#8531;",
    "2/3": "&#8532;",
    "1/5": "&#8533;",
    "2/5": "&#8534;",
    "3/5": "&#8535;",
    "4/5": "&#8536;",
    "1/6": "&#8537;",
    "5/6": "&#8538;",
    "1/8": "&#8539;",
    "3/8": "&#8540;",
    "5/8": "&#8541;",
    "7/8": "&#8542;"
}

ARR = {
    '-->': "&rarr;",
    '<--': "&larr;",
    '<-->': "&harr;"
}


class SmartSymbolsPattern(HtmlInlineProcessor):
    """Smart symbols patterns handler."""

    def __init__(self, pattern, replace, md):
        """Setup replace pattern."""

        super(SmartSymbolsPattern, self).__init__(pattern, md)
        self.replace = replace

    def handleMatch(self, m, data):
        """Replace symbol."""

        return self.md.htmlStash.store(
            m.expand(self.replace(m) if callable(self.replace) else self.replace),
        ), m.start(0), m.end(0)


class SmartSymbolsExtension(Extension):
    """Smart Symbols extension."""

    def __init__(self, *args, **kwargs):
        """Setup config of which symbols are enabled."""

        self.config = {
            'trademark': [True, 'Trademark'],
            'copyright': [True, 'Copyright'],
            'registered': [True, 'Registered'],
            'plusminus': [True, 'Plus/Minus'],
            'arrows': [True, 'Arrows'],
            'notequal': [True, 'Not Equal'],
            'fractions': [True, 'Fractions'],
            'ordinal_numbers': [True, 'Ordinal Numbers'],
            'care_of': [True, 'Care/of']
        }
        super(SmartSymbolsExtension, self).__init__(*args, **kwargs)

    def add_pattern(self, patterns, md):
        """Construct the inline symbol pattern."""

        self.patterns.register(SmartSymbolsPattern(patterns[1], patterns[2], md), patterns[0], 30)

    def extendMarkdown(self, md):
        """Create a dict of inline replace patterns and add to the tree processor."""

        configs = self.getConfigs()
        self.patterns = Registry()

        for k, v in REPL.items():
            if configs[k]:
                self.add_pattern(v, md)

        inline_processor = treeprocessors.InlineProcessor(md)
        inline_processor.inlinePatterns = self.patterns
        md.treeprocessors.register(inline_processor, "smart-symbols", 2.1)


def makeExtension(*args, **kwargs):
    """Return extension."""

    return SmartSymbolsExtension(*args, **kwargs)
